/*
 * Export MIPS Creator CI20 HDMI display operations as a server.
 *
 * Copyright (C) 2018, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/cpm-client.h>
#include <l4/devices/hdmi-jz4780.h>
#include <l4/devices/lcd-jz4740-config.h>
#include <l4/devices/lcd-jz4740-panel.h>
#include <l4/devices/panel-loader.h>
#include <l4/devices/memory.h>

#include <l4/cxx/ipc_server>
#include <l4/re/env>
#include <l4/re/util/object_registry>
#include <l4/sys/capability>
#include <l4/sys/factory>
#include <l4/sys/icu.h>

#include "display-ops.h"

/* Virtual addresses for the HDMI register block plus interrupt capability. */

static l4_addr_t hdmi_virt_base = 0, hdmi_virt_base_end = 0;
static l4_uint32_t hdmi_irq_start, hdmi_irq_end;
static L4::Cap<L4::Irq> hdmi_irq;
static L4::Cap<L4::Icu> icu;

/* CPM device abstractions. */

static Hdmi_jz4780_chip *hdmi = 0;
static L4::Cap<Cpm_device_interface> cpm_device;



static int setup_memory(void)
{
  if (get_memory("jz4780-hdmi", &hdmi_virt_base, &hdmi_virt_base_end))
    return 1;

  if (get_irq("jz4780-hdmi", &hdmi_irq_start, &hdmi_irq_end))
    return 1;

  /* Obtain a reference to the CPM device. */

  cpm_device = L4Re::Env::env()->get_cap<Cpm_device_interface>("cpm");
  if (!cpm_device.is_valid()) return 1;

  /* Start the HDMI peripheral. */

  cpm_device->stop_hdmi();
  cpm_device->set_hdmi_frequency(27000000);
  cpm_device->start_hdmi();

  /* Load the panel data from the configured library. */

  struct Jz4740_lcd_panel *panel = (struct Jz4740_lcd_panel *) load_panel();
  if (!panel) return 1;

  /* Obtain access to the HDMI interrupt using the ICU. */

  icu = L4Re::Env::env()->get_cap<L4::Icu>("icu");

  hdmi_irq = L4Re::Util::cap_alloc.alloc<L4::Irq>();
  L4Re::Env::env()->factory()->create(hdmi_irq);

  if (!hdmi_irq.is_valid() || !icu.is_valid()) return 1;

  /* Bind the IRQ object to the interrupt in this thread. */

  int err = l4_error(icu->bind(hdmi_irq_start, hdmi_irq));
  if (err) return err;

  hdmi_irq->bind_thread(L4Re::Env::env()->main_thread(), 0);

  /* Obtain an abstraction for the HDMI peripheral. */

  hdmi = new Hdmi_jz4780_chip(hdmi_virt_base, hdmi_virt_base_end, hdmi_irq.cap(), panel);

  return 0;
}



/* Display device. */

class Display_device_server : public L4::Server_object_t<L4::Kobject>
{
public:
  /* Dispatch incoming requests. */

  int dispatch(l4_umword_t obj, L4::Ipc::Iostream &ios)
  {
    l4_msgtag_t tag;

    (void) obj;
    ios >> tag;

    switch (tag.label())
    {
      case Display_op_disable:
        disable();
        return L4_EOK;

      case Display_op_enable:
        enable();
        return L4_EOK;

      default:
        return -L4_EBADPROTO;
    }
  }

  /* Switch the display off. */

  void disable(void)
  {
  }

  /* Switch the display on. */

  void enable(void)
  {
    uint32_t frequency;

    // NOTE: Should test to see if it is connected.

    if (!cpm_device->get_lcd_pixel_frequency(&frequency))
    {
      hdmi->enable(frequency);
    }
  }
};

static L4Re::Util::Registry_server<> server;



int main(void)
{
  if (setup_memory()) return 1;

  /* Initialise and register a new server object. */

  Display_device_server server_obj;
  server.registry()->register_obj(&server_obj, "display");

  /* Enter the IPC server loop. */

  server.loop();
  return 0;
}
